// @TITLE "ActiveSet.h: class ActiveSet"
//
//  This class implements the active set.  There is only one active
// set, which determines which VPs are active at any given time.  The
// ifp() macro, however, can temporarily restrict the active set to the
// subset where the condition is true, and then to where the condition is
// false.  These can be nested, so the active set actually depends on an
// active stack to manage a stack of active sets.
//   There are no methods of this class that should be
// called directly by the end user -- they are meant to be called by the macros
// or by the Collection/Vector/Matrix functions.
//
// David Kotz 1994

// $Id: ActiveSet.h,v 2.16 95/12/17 00:08:01 dfk release Locker: dfk $

#ifndef ACTIVESET_H
#define ACTIVESET_H

#include <stdlib.h>
#include "macros.h"
// class booleanCollection must already be defined.

class ActiveStack;		// contents not important here

class ActiveSet {
public:
    ActiveSet();
    ~ActiveSet();

    // these are used by ifp(), and should not be used for anything else
    void restrict(CONST booleanCollection& condition); // start parallel block
    boolean loop(void);			    // and continue it...
    boolean which(void);		    // then (true) or else (false)

    // these are used by Collection, Vector, etc; not for general use
    boolean isActive(CONST int vp) CONST     // for vectors
      { return (everywhere || current == NULL || (*current)(vp)); }
    boolean isActive(CONST int vr, CONST int vc) CONST  // for matrices
      { return (everywhere || current == NULL || (*current)(vr,vc)); }
    boolean shapeOK(CONST Shape& s) CONST 
      // check a shape to see if we are allowed to use it in this context
      { return (current == NULL || s == current->shape); }

    char *shapeName(void) CONST {
	if (current == NULL)
	  return("ALL");
	else
	  return(current->shape.name());
    }

private:
    booleanCollection *current; // which processors are active
    boolean everywhere;		  // used internally to override active set
    int whichLoop;		  // which iteration in ifp()
    ActiveStack *enclosing;	  // stack of states; top is the enclosing scope
};

extern ActiveSet Active;	// the current active set object

//////////////////////////////////////////
// A parallel conditional statement, complete with "else" clause:
//   ifp (condition) {
//     then clause...
//   } else {
//     else clause...
//   }
// The brackets are only necessary when they would be in normal C.
// A collection condition should be provided.  Those processors that are
// already active and where the condition is true will execute the THEN
// part of the ifp, and those who are either already inactive or where
// the condition is false will execute the ELSE part of the ifp.  Note
// that both parts will be executed once, so any scalar expressions will
// be evaluated once.  
//   Implementation: loop() always returns true twice, and false once,
// so the FOR loop goes twice.  which() returns true the first time,
// and false the second time, with a different set of VPs active each time.
// Nesting is ok.
#define ifp(condition) \
    for (Active.restrict(condition); Active.loop(); ) \
    	if (Active.which())

#endif // ACTIVESET_H
