// @TITLE "Matrix.cc: Matrix classes"
//
// Matrix classes; included by {int,char,float,double,boolean}.cc.  
//
// David Kotz 1994

// $Id: Matrix.cc,v 2.16 95/12/17 00:07:57 dfk release Locker: dfk $

/////////////////////////////////////////////////////////////////////
// Other constructors and some operators are declared inline in Matrix.h

Matrix::Matrix(CONST intMatrix& M) : Collection(M) {}
Matrix::Matrix(CONST charMatrix& M) : Collection(M) {}
Matrix::Matrix(CONST floatMatrix& M) : Collection(M) {}
Matrix::Matrix(CONST doubleMatrix& M) : Collection(M) {}
Matrix::Matrix(CONST booleanMatrix& M) : Collection(M) {}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::init: private function to initialize matrix"
//  This variant initializes the array to a function of the indices.
// Note that all entries are initialized, regardless of active set.

void Matrix::init( Element (*f)(CONST int index, CONST int index2) )
{
    // now initialize v[] using the function
    for (int r = 0; r < shape.dim1; r++)
      for (int c = 0; c < shape.dim2; c++)
	(*this)(r,c) = f(r, c);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::init: private function to initialize matrix"
//  Given an array of integers, initialize this object.
// Note that all entries are initialized, regardless of active set.

void Matrix::init(CONST Element **array)
{
    // now initialize v[] using the array
    for (int r = 0; r < shape.dim1; r++)
      for (int c = 0; c < shape.dim2; c++)
	(*this)(r,c) = array[r][c];
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::load: load matrix from array"
//    copy the plain array into the matrix
//  the array better have be large enough

void Matrix::load(CONST Element **CONST array)
{
    INVARIANT2(array != NULL, "NULL array passed to Matrix::load.");
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int r = 0; r < shape.dim1; r++)
      for (int c = 0; c < shape.dim2; c++)
	if (Active.isActive(r,c))
	  (*this)(r,c) = array[r][c];
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::store: store matrix into array"
//    copy the matrix into a plain integer array.
//  the array better have enough space

void Matrix::store(Element **CONST array) CONST
{
    INVARIANT2(array != NULL, "NULL array passed to Matrix::store.");
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int r = 0; r < shape.dim1; r++)
      for (int c = 0; c < shape.dim2; c++)
	if (Active.isActive(r,c))
	  array[r][c] = (*this)(r,c);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::extract: extract a vector from matrix"
//    Pull a row or column out of the array, returning a vector.
// This is used by the Slice class.  The rules of conformity are a little 
// different than in most situations: the
// active set may conform to EITHER the vector or the matrix, and the
// assignment is conditioned on the appropriate index space.

Vector Matrix::extract(CONST SliceType type, CONST int index) CONST
{
    switch(type) {
	case rowSlice: {
	    Vector result(Shape('V', shape.dim2));

	    if (Active.shapeOK(result.shape)) {	// conforms to vector
		for (int r = index, c = 0; c < shape.dim2; c++)
		  if (Active.isActive(c))
		    result(c) = (*this)(r,c);
		  else
		    result(c) = false;
	    } else if (Active.shapeOK(shape)) {	// conforms to matrix
		for (int r = index, c = 0; c < shape.dim2; c++)
		  if (Active.isActive(r,c))
		    result(c) = (*this)(r,c);
		  else
		    result(c) = false;
	    } else
	      INVARIANT4(false, // check conformity with active set
			 "Cannot extract row slice from matrix shape %s in context shape %s",
			 shape.name(), Active.shapeName());
	    
	    return (result);
	}
    	case colSlice: {
	    Vector result(Shape('V', shape.dim1));
			   
	    if (Active.shapeOK(result.shape)) { // conforms to vector
		for (int c = index, r = 0; r < shape.dim1; r++)
		  if (Active.isActive(r))
		    result(r) = (*this)(r,c);
		  else
		    result(c) = false;
	    } else if (Active.shapeOK(shape)) { // conforms to matrix
		for (int c = index, r = 0; r < shape.dim1; r++)
		  if (Active.isActive(r,c))
		    result(r) = (*this)(r,c);
		  else
		    result(c) = false;
	    } else {
		INVARIANT4(false, // check conformity with active set
			   "Cannot extract column slice from matrix shape %s in context shape %s",
			   shape.name(), Active.shapeName());
	    }
	    return (result);
	}
	default: {
	    INVARIANT2(false, "(internal) unknown slice type.");
	    Vector result(Shape('V', 1)); // not reached; to keep compiler happy
	    return(result);
	}
    }
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Matrix::update: update the matrix from a vector"
//    Copy a row or column into the array, given a vector.
// This is used by the Slice class.  The rules of conformity are a little 
// different than in most situations.  Certainly, the vector V must have
// the same shape as the corresponding slice of the Matrix.  But, the
// active set may conform to EITHER the vector or the matrix, and the
// assignment is conditioned on the appropriate index space.

void Matrix::update(CONST SliceType type, CONST int index, CONST Vector& V)
{
    switch(type) {
	case rowSlice: {
	    Shape correct('V', shape.dim2);
	    INVARIANT4(V.shape == correct,
		       "You can't update a row (shape %s) from vector (shape %s)",
		       correct.name(), V.shape.name());
	    
	    if (Active.shapeOK(V.shape)) { // conforms to vector
		for (int r = index, c = 0; c < shape.dim2; c++)
		  if (Active.isActive(c))
		    (*this)(r,c) = V(c);
	    } else if (Active.shapeOK(shape)) {	// conforms to matrix
		for (int r = index, c = 0; c < shape.dim2; c++)
		  if (Active.isActive(r,c))
		    (*this)(r,c) = V(c);
	    } else 
	      INVARIANT4(false, // check conformity with active set
			 "Cannot update row slice of matrix shape %s in context shape %s",
			 shape.name(), Active.shapeName());
	    
	    break;
	}
        case colSlice: {
	    Shape correct('V', shape.dim1);
	    INVARIANT4(V.shape == correct,
		       "You can't update a column (shape %s) from vector (shape %s)",
		       correct.name(), V.shape.name());
			   
	    if (Active.shapeOK(V.shape)) { // conforms to vector
		for (int c = index, r = 0; r < shape.dim1; r++)
		  if (Active.isActive(r))
		    (*this)(r,c) = V(r);
	    } else if (Active.shapeOK(shape)) {	// conforms to matrix
		for (int c = index, r = 0; r < shape.dim1; r++)
		  if (Active.isActive(r,c))
		    (*this)(r,c) = V(r);
	    } else 
	      INVARIANT4(false, // check conformity with active set
			 "Cannot update column slice of matrix shape %s in context shape %s",
			 shape.name(), Active.shapeName());
	    
	    
	    break;
	}
        default: {
	    INVARIANT2(false, "(internal) unknown slice type.");
	    break;
	}
    }
}

//////////////// Friends of the Matrix class ///////////////////

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "or_scan_rows(Matrix)"
//  Apply an orring scan to each row of this matrix, returning a new matrix.
// Note this is a boolean or, not a bitwise or.

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Matrix or_scan_rows(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++) {
	Element m = false;
	
	for (int c = 0; c < M.shape.dim2; c++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = (m || M(r,c));
	  } else
	    result(r,c) =false;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "or_scan_cols(Matrix)"
//  Apply an orring scan to each col of this matrix, returning a new matrix.
// Note this is a boolean or, not a bitwise or.

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Matrix or_scan_cols(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int c = 0; c < M.shape.dim2; c++) {
	Element m = false;
	
	for (int r = 0; r < M.shape.dim1; r++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = (m || M(r,c));
	  } else
	    result(r,c) = false;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "and_scan_rows(Matrix)"
//  Apply an anding scan to each row of this matrix, returning a new matrix.
// Note this is a boolean and, not a bitwise and.

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Matrix and_scan_rows(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++) {
	Element m = true;
	
	for (int c = 0; c < M.shape.dim2; c++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = (m && M(r,c));
	  } else
	    result(r,c) = false;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "and_scan_cols(Matrix)"
//  Apply an anding scan to each col of this matrix, returning a new matrix.
// Note this is a boolean and, not a bitwise and.

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Matrix and_scan_cols(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int c = 0; c < M.shape.dim2; c++) {
	Element m = true;
	
	for (int r = 0; r < M.shape.dim1; r++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m =  (m && M(r,c));
	  } else
	    result(r,c) = false;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "plus_scan_rows(Matrix)"
//  Apply a summing scan to each row of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support plus
Matrix plus_scan_rows(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++) {
	Element m = 0;
	
	for (int c = 0; c < M.shape.dim2; c++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = m + M(r,c);
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "plus_scan_cols(Matrix)"
//  Apply a summing scan to each col of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support plus
Matrix plus_scan_cols(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int c = 0; c < M.shape.dim2; c++) {
	Element m = 0;
	
	for (int r = 0; r < M.shape.dim1; r++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = m + M(r,c);
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "max_scan_rows(Matrix)"
//  Apply a maxing scan to each row of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max
Matrix max_scan_rows(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++) {
	Element m = Smallest();
	
	for (int c = 0; c < M.shape.dim2; c++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = max(m, M(r,c));
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "max_scan_cols(Matrix)"
//  Apply a maxing scan to each col of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max
Matrix max_scan_cols(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int c = 0; c < M.shape.dim2; c++) {
	Element m = Smallest();
	
	for (int r = 0; r < M.shape.dim1; r++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = max(m, M(r,c));
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "min_scan_rows(Matrix)"
//  Apply a minning scan to each row of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support min
Matrix min_scan_rows(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++) {
	Element m = Biggest();
	
	for (int c = 0; c < M.shape.dim2; c++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = min(m, M(r,c));
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "min_scan_cols(Matrix)"
//  Apply a minning scan to each col of this matrix, returning a new matrix.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support min
Matrix min_scan_cols(CONST Matrix& M)
{
    Matrix result(M.shape);	// uninitialized
    
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int c = 0; c < M.shape.dim2; c++) {
	Element m = Biggest();
	
	for (int r = 0; r < M.shape.dim1; r++)
	  if (Active.isActive(r,c)) {
	      result(r,c) = m;
	      m = min(m, M(r,c));
	  } else
	    result(r,c) = 0;
    }
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "shift(Matrix,dr,dc): shift the matrix"
//   Shift the given matrix by the given offsets.  Any positive or negative
// offsets will do, but anything larger than +-M.shape.dim1 (or M.shape.dim2)
// is silly. Positive offsets push things to the right (toward larger indices).
// The vacated spaces are filled with zeroes (may want more flexibility later).
// Inactive places in the matrix are untouched. 
//   Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is shifted.
// [Alternatively, it could shift only the active elements,
// or shift them among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix shift(CONST Matrix& M, CONST int deltaR, CONST int deltaC)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    if (deltaR == 0 && deltaC == 0)
      return M;			// zero shift is a no-op
    else  {
	Matrix result(M.shape.dim1, M.shape.dim2, (Element)0);
	
	int cmin = max(0,              0              - deltaC);
	int cmax = min(M.shape.dim2, M.shape.dim2 - deltaC);
	int rmin = max(0,              0              - deltaR);
	int rmax = min(M.shape.dim1, M.shape.dim1 - deltaR);
	
	// now fill in those places that are not zero
	// move position r to position r+deltaR
	//               c to position c+deltaC
	for (int r = rmin; r < rmax; r++)
	  for (int c = cmin; c < cmax; c++)
	    result((r + deltaR), (c + deltaC)) = M(r,c);
	
	return result;
    }
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "shift_rows(Matrix,cv): shift each row of the matrix"
//   Shift each row of the given matrix by a different offset, specified
// in a column vector.  Any positive or negative
// offsets will do, but anything larger than +-M.shape.dim2
// is silly. Positive offsets push things to the right (toward larger indices).
// The vacated spaces are filled with zeroes (may want more flexibility later).
// Inactive places in the matrix are untouched. 
//   Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is shifted.
// [Alternatively, it could shift only the active elements,
// or shift them among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix shift_rows(CONST Matrix& M, CONST intVector& offsets)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    Matrix result(M.shape.dim1, M.shape.dim2, (Element)0);
    
    // now fill in those places that are not zero
    // move position c to position c+offset(r)
    for (int r = 0; r < M.shape.dim1; r++) {
	int offset = offsets(r); // row r's offset
	int cmin = max(0,            0            - offset);
	int cmax = min(M.shape.dim2, M.shape.dim2 - offset);
	
	for (int c = cmin; c < cmax; c++)
	  result(r, (c + offset)) = M(r,c);
    }
    
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "shift_cols(Matrix,rv): shift each column of the matrix"
//   Shift each column of the given matrix by a different offset, specified
// in a row vector.  Any positive or negative
// offsets will do, but anything larger than +-M.shape.dim1
// is silly. Positive offsets push things down (toward larger indices).
// The vacated spaces are filled with zeroes (may want more flexibility later).
// Inactive places in the matrix are untouched. 
//   Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is shifted.
// [Alternatively, it could shift only the active elements,
// or shift them among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix shift_cols(CONST Matrix& M, CONST intVector& offsets)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    Matrix result(M.shape.dim1, M.shape.dim2, (Element)0);
    
    // now fill in those places that are not zero
    //               r to position r+offset(c)
    for (int c = 0; c < M.shape.dim2; c++) {
	int offset = offsets(c); // row r's offset
	int rmin = max(           0,            0 - offset);
	int rmax = min(M.shape.dim1, M.shape.dim1 - offset);
	
	for (int r = rmin; r < rmax; r++)
	  result((r + offset), c) = M(r,c);
    }
    
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "rotate(Matrix,dr,dc): rotate the matrix"
//   Rotate the given matrix by the given offsets.  Any positive or negative
// offset will do, but anything larger than +-M.shape.dim1 (or M.shape.dim2)
// is redundant. // Positive offsets push things to the right
// (toward larger indices).
//    Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is rotated deltas 
// places to the right.  [Alternatively, it could rotate only the active 
// elements, among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix rotate(CONST Matrix& M, CONST int deltaR, CONST int deltaC)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    int offsetR = deltaR; // local, changable copy
    int offsetC = deltaC; // local, changable copy
    
    offsetR %= M.shape.dim1; // eliminate extra full rotations
    offsetR += M.shape.dim1; // convert negative to positive
    offsetR %= M.shape.dim1; // ie, now it is always a right rotation
    
    offsetC %= M.shape.dim2; // eliminate extra full rotations
    offsetC += M.shape.dim2; // convert negative to positive
    offsetC %= M.shape.dim2; // ie, now it is always a right rotation
    
    if (offsetR == 0 && offsetC == 0)
      return M;	// zero rotation is a no-op
    else {
	Matrix result(M.shape); // uninitialized
	
	// move position r to position r+offsetR, mod the matrix rows
	//               c to position c+offsetC, mod the matrix cols
	for (int r = 0; r < M.shape.dim1; r++)
	  for (int c = 0; c < M.shape.dim2; c++)
	    result((r + offsetR) % M.shape.dim1,
		   (c + offsetC) % M.shape.dim2) = M(r,c);
	
	return result;
    }
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "rotate_rows(Matrix,cv): rotate the matrix; each row differently"
//   Rotate the given matrix, by rotating each row by the amount given in the 
// column vector.  Any positive or negative offset will do, but anything larger 
// than +-M.shape.dim2 is redundant. 
// Positive offsets push things to the right (toward larger indices).
//    Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is rotated deltas 
// places to the right.  [Alternatively, it could rotate only the active 
// elements, among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix rotate_rows(CONST Matrix& M, CONST intVector& offsets)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    INVARIANT4(offsets.shape == Shape('V', M.shape.dim1),
	       "Rotate_rows: offset shape %s is not a column vector for matrix %s",
	       offsets.shape.name(), M.shape.name());
    
    Matrix result(M.shape); // uninitialized
    
    // in each row, move col c to position c+offsets, mod the matrix cols
    for (int r = 0; r < M.shape.dim1; r++) {
	int offset = offsets(r); // row r's offset
	offset %= M.shape.dim2; // eliminate extra full rotations
	offset += M.shape.dim2; // convert negative to positive
	offset %= M.shape.dim2; // ie, now it is always a right rotation
	
	for (int c = 0; c < M.shape.dim2; c++)
	  result(r, (c + offset) % M.shape.dim2) = M(r,c);
    }
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "rotate_cols(Matrix,rv): rotate the matrix; each column differently"
//   Rotate the given matrix, by rotating each column by the amount given in the 
// row vector.  Any positive or negative offset will do, but anything 
// larger than +-M.shape.dim1 is redundant. 
// Positive offsets push things to the bottom (toward larger indices).
//    Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole matrix is rotated deltas 
// places to the right.  [Alternatively, it could rotate only the active 
// elements, among the active positions.  So far, I've
// found the semantics here to be more useful.]

Matrix rotate_cols(CONST Matrix& M, CONST intVector& offsets)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    INVARIANT4(offsets.shape == Shape('V', M.shape.dim2),
	       "Rotate_rows: offset shape %s is not a row vector for matrix %s",
	       offsets.shape.name(), M.shape.name());
    
    Matrix result(M.shape); // uninitialized
    
    // in each column, move row r to position r+offsets, mod the matrix rows
    for (int c = 0; c < M.shape.dim2; c++) {
	int offset = offsets(c); // col c's offset
	offset %= M.shape.dim1; // eliminate extra full rotations
	offset += M.shape.dim1; // convert negative to positive
	offset %= M.shape.dim1; // ie, now it is always a right rotation
	
	for (int r = 0; r < M.shape.dim1; r++) 
	  result((r + offset) % M.shape.dim1, c) = M(r,c);
    }
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "ostream operator <<: print operator"
//   Print matrix A to the given output stream, in row-major order.
// Only active elements are printed.
// A newline is appended after each row;  elements are tab-separated.

ostream& operator<< (ostream& out, CONST Matrix& M)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    boolean startline;
    
    for (int r = 0; r < M.shape.dim1; r++) {
	startline = true;	// we're starting a new line

	for (int c = 0; c < M.shape.dim2; c++) {
	    if (Active.isActive(r,c)) {
#ifndef CHAR_ELEMENTS
		if (startline)
		  startline = false;
		else
		  out << "\t";	// insert tab between entries
#else
		startline = false;
#endif				// CHAR_ELEMENTS

		out << M(r,c);
	    }
	}
	if (!startline)
	  out << endl;
    }
    return(out);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "istream operator >>: read operator"
//   Read matrix M from the given input stream.  Only active elements
// are read. No newline is expected.
// Elements should be whitespace-separated,
// and provided in row-major order.

istream& operator>> (istream& in, Matrix& M)
{
    INVARIANT4(Active.shapeOK(M.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       M.shape.name(), Active.shapeName());
    
    for (int r = 0; r < M.shape.dim1; r++)
      for (int c = 0; c < M.shape.dim2; c++)
	if (Active.isActive(r,c))
	  in >> M(r,c);
    return(in);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "outer(): outer product of two vectors"
// compute outer product A * B
// if A has p elements, and B has q elements, the result is a matrix
// of p rows and q columns.  Only those locations in the result matrix that
// correspond to active VPs will be computed.

#ifndef BOOLEAN_ELEMENTS	// disallow arithmetic on booleans
Matrix outer(CONST Vector& A, CONST Vector& B)
{
    Matrix result(Shape('M', A.shape.dim1, B.shape.dim1));
    
    INVARIANT4(Active.shapeOK(result.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       result.shape.name(), Active.shapeName());
    
    for (int r = 0; r < A.shape.dim1; r++)
      for (int c = 0; c < B.shape.dim1; c++)
	if (Active.isActive(r,c))
	  result(r,c) = A(r) * B(c);
	else
	  result(r,c) = 0;
    
    return(result);
}
#endif

/////////////////////////////////////////////////////////////////////
