// @TITLE "Matrix.h: class Matrix"
//
// included by dapple.h -- do not include this directly!
//
// Matrix class.  A Matrix is a subclass of Collection, so we can inherit
// many of the elementwise operators.  Thus, we need only add functions that
// make sense on Matrices; in particular, functions that involve the
// understanding that the elements of the collection have some order, 
// and each has some index.  This includes 
// input and output, and some new constructors to initialize from an array or
// a function of the indices.
//
// David Kotz 1994

// $Id: Matrix.h,v 2.16 95/12/17 00:07:49 dfk release Locker: dfk $

class Matrix : public Collection {

    friend class Slice;
    friend class HalfSlice;

    // they're all friends
#if !defined(INT_ELEMENTS)
    friend class intMatrix;
#endif
#if !defined(CHAR_ELEMENTS)
    friend class charMatrix;
#endif
#if !defined(FLOAT_ELEMENTS)
    friend class floatMatrix;
#endif
#if !defined(DOUBLE_ELEMENTS)
    friend class doubleMatrix;
#endif
#if !defined(BOOLEAN_ELEMENTS)
    friend class booleanMatrix;
#endif

public: 
    // constructors and destructors
    Matrix(CONST int rows, CONST int cols)
      : Collection(Shape('M', rows, cols)) {};
    Matrix(CONST int rows, CONST int cols, CONST Element x)
      : Collection(Shape('M', rows, cols), x) {}
    Matrix(CONST int rows, CONST int cols, CONST Element **d)
      : Collection(Shape('M', rows, cols)) { init(d); }
    Matrix(CONST int rows, CONST int cols, Element (*f)(CONST int r, CONST int c))
      : Collection(Shape('M', rows, cols)) { init(f); }
    Matrix(CONST intMatrix& M);	// similar to the above
    Matrix(CONST charMatrix& M); // but defined in Matrix.cc
    Matrix(CONST floatMatrix& M);
    Matrix(CONST doubleMatrix& M);
    Matrix(CONST booleanMatrix& M);

public:		// internal; should be private! but strictly speaking won't work that way.
    Matrix(CONST Shape& s) : Collection(s) {}
    
public:
    // overloaded operators
    Matrix& operator=(CONST Matrix& B)   // A = B
      { this->OP_Equals(B); return *this; }
    Matrix& operator=(CONST Element x)   // A = x
      { this->OP_Equals(x); return *this; }
#ifndef BOOLEAN_ELEMENTS   // don't allow arithmetic on booleans
    Matrix& operator+=(CONST Matrix& B)   // A += B
      { this->OP_PlusEquals(B); return(*this); }
    Matrix& operator+=(CONST Element x)   // A += x
      { this->OP_PlusEquals(x); return(*this); }
    Matrix& operator-=(CONST Matrix& B)   // A -= B
      { this->OP_MinusEquals(B); return(*this); }
    Matrix& operator-=(CONST Element x)   // A -= x
      { this->OP_MinusEquals(x); return(*this); }
    Matrix& operator*=(CONST Matrix& B)   // A *= B
      { this->OP_TimesEquals(B); return(*this); }
    Matrix& operator*=(CONST Element x)   // A *= x
      { this->OP_TimesEquals(x); return(*this); }
    Matrix& operator/=(CONST Matrix& B)   // A /= B
      { this->OP_DivideEquals(B); return(*this); }
    Matrix& operator/=(CONST Element x)   // A /= x
      { this->OP_DivideEquals(x); return(*this); }
# ifdef INT_ELEMENTS // not for char, float, double, or boolean
    Matrix& operator%=(CONST Matrix& B)   // A %= B
      { this->OP_ModEquals(B); return(*this); }
    Matrix& operator%=(CONST Element x)   // A %= x
      { this->OP_ModEquals(x); return(*this); }
# endif
    Matrix  operator++(void)  	         // ++A
      { this->OP_PlusPlus(); return(*this); }
    Matrix  operator++(int zero)   // A++
      { Matrix result(*this); this->OP_PlusPlus(zero); return(result); }
    Matrix  operator--(void)  	         // --A
      { this->OP_MinusMinus(); return(*this); }
    Matrix  operator--(int zero)   // A--
      { Matrix result(*this); this->OP_MinusMinus(zero); return(result); }

    friend Matrix       operator+(CONST Matrix& A)  	  	      // +A 
      {	return (A); }
    friend Matrix       operator-(CONST Matrix& A) 		       // -A 
      { Matrix result(A.shape);	OP_Minus(result, A); return(result); }
    friend Matrix       operator+(CONST Matrix& A, CONST Matrix& B)   // A + B
      {	Matrix result(A.shape); OP_Plus(result, A, B); return(result); }
    friend Matrix       operator+(CONST Matrix& A, CONST Element x)   // A + x
      {	Matrix result(A.shape); OP_Plus(result, A, x); return(result); }
    friend Matrix       operator+(CONST Element x, CONST Matrix& B)   // x + B
      {	Matrix result(B.shape); OP_Plus(result, x, B); return(result); }
    friend Matrix       operator-(CONST Matrix& A, CONST Matrix& B)   // A - B
      { Matrix result(A.shape); OP_Minus(result, A, B); return(result); }
    friend Matrix       operator-(CONST Matrix& A, CONST Element x)   // A - x
      { Matrix result(A.shape); OP_Minus(result, A, x); return(result); }
    friend Matrix       operator-(CONST Element x, CONST Matrix& B)   // x - B
      { Matrix result(B.shape); OP_Minus(result, x, B); return(result); }
    friend Matrix       operator*(CONST Matrix& A, CONST Matrix& B)   // A * B
      { Matrix result(A.shape); OP_Times(result, A, B); return(result); }
    friend Matrix       operator*(CONST Matrix& A, CONST Element x)   // A * x
      { Matrix result(A.shape); OP_Times(result, A, x); return(result); }
    friend Matrix       operator*(CONST Element x, CONST Matrix& B)   // x * B
      { Matrix result(B.shape); OP_Times(result, x, B); return(result); }
    friend Matrix       operator/(CONST Matrix& A, CONST Matrix& B)   // A / B
      { Matrix result(A.shape); OP_Divide(result, A, B); return(result); }
    friend Matrix       operator/(CONST Matrix& A, CONST Element x)   // A / x
      { Matrix result(A.shape); OP_Divide(result, A, x); return(result); }
    friend Matrix       operator/(CONST Element x, CONST Matrix& B)   // x / B
      { Matrix result(B.shape); OP_Divide(result, x, B); return(result); }
# ifdef INT_ELEMENTS			       // not for char, float, double, or boolean
    friend Matrix       operator%(CONST Matrix& A, CONST Matrix& B)   // A % B
      { Matrix result(A.shape); OP_Mod(result, A, B); return(result); }
    friend Matrix       operator%(CONST Matrix& A, CONST Element x)   // A % x
      { Matrix result(A.shape); OP_Mod(result, A, x); return(result); }
    friend Matrix       operator%(CONST Element x, CONST Matrix& B)   // x % B
      { Matrix result(B.shape); OP_Mod(result, x, B); return(result); }
# endif // INT_ELEMENTS

    friend booleanMatrix operator<(CONST Matrix& A, CONST Matrix& B)   // A < B
      { booleanMatrix result(A.shape); OP_LT(result, A, B); return(result); }
    friend booleanMatrix operator<(CONST Matrix& A, CONST Element x)   // A < x
      { booleanMatrix result(A.shape); OP_LT(result, A, x); return(result); }
    friend booleanMatrix operator<(CONST Element x, CONST Matrix& B)   // x < B
      { booleanMatrix result(B.shape); OP_LT(result, x, B); return(result); }
    friend booleanMatrix operator>(CONST Matrix& A, CONST Matrix& B)   // A > B
      { booleanMatrix result(A.shape); OP_GT(result, A, B); return(result); }
    friend booleanMatrix operator>(CONST Matrix& A, CONST Element x)   // A > x
      { booleanMatrix result(A.shape); OP_GT(result, A, x); return(result); }
    friend booleanMatrix operator>(CONST Element x, CONST Matrix& B)   // x > B
      { booleanMatrix result(B.shape); OP_GT(result, x, B); return(result); }
    friend booleanMatrix operator<=(CONST Matrix& A, CONST Matrix& B)   // A <= B
      { booleanMatrix result(A.shape); OP_LE(result, A, B); return(result); }
    friend booleanMatrix operator<=(CONST Matrix& A, CONST Element x)  	// a <= x
      { booleanMatrix result(A.shape); OP_LE(result, A, x); return(result); }
    friend booleanMatrix operator<=(CONST Element x, CONST Matrix& B)  	// x <= B
      { booleanMatrix result(B.shape); OP_LE(result, x, B); return(result); }
    friend booleanMatrix operator>=(CONST Matrix& A, CONST Matrix& B)   // A >= B
      { booleanMatrix result(A.shape); OP_GE(result, A, B); return(result); }
    friend booleanMatrix operator>=(CONST Matrix& A, CONST Element x)  	// A >= x
      { booleanMatrix result(A.shape); OP_GE(result, A, x); return(result); }
    friend booleanMatrix operator>=(CONST Element x, CONST Matrix& B)  	// x >= B
      { booleanMatrix result(B.shape); OP_GE(result, x, B); return(result); }
#endif // BOOLEAN_ELEMENTS
    friend booleanMatrix operator==(CONST Matrix& A, CONST Matrix& B)   // A == B
      { booleanMatrix result(A.shape); OP_EQ(result, A, B); return(result); }
    friend booleanMatrix operator==(CONST Matrix& A, CONST Element x)  	// A == x
      { booleanMatrix result(A.shape); OP_EQ(result, A, x); return(result); }
    friend booleanMatrix operator==(CONST Element x, CONST Matrix& B)  	// x == B
      { booleanMatrix result(B.shape); OP_EQ(result, x, B); return(result); }
    friend booleanMatrix operator!=(CONST Matrix& A, CONST Matrix& B)   // A != B
      { booleanMatrix result(A.shape); OP_NE(result, A, B); return(result); }
    friend booleanMatrix operator!=(CONST Matrix& A, CONST Element x)  	// A != x
      { booleanMatrix result(A.shape); OP_NE(result, A, x); return(result); }
    friend booleanMatrix operator!=(CONST Element x, CONST Matrix& B)  	// x != B
      { booleanMatrix result(B.shape); OP_NE(result, x, B); return(result); }

#ifdef BOOLEAN_ELEMENTS   // only for booleans
    friend booleanMatrix operator&&(CONST Matrix& A, CONST Matrix& B)   // A && B
      { booleanMatrix result(A.shape); OP_AND(result, A, B); return(result); }
    friend booleanMatrix operator&&(CONST Matrix& A, CONST Element x)  	// A && x
      { booleanMatrix result(A.shape); OP_AND(result, A, x); return(result); }
    friend booleanMatrix operator&&(CONST Element x, CONST Matrix& B)  	// x && B
      { booleanMatrix result(B.shape); OP_AND(result, x, B); return(result); }
    friend booleanMatrix operator||(CONST Matrix& A, CONST Matrix& B)   // A || B
      { booleanMatrix result(A.shape); OP_OR(result, A, B); return(result); }
    friend booleanMatrix operator||(CONST Matrix& A, CONST Element x)  	// A || x
      { booleanMatrix result(A.shape); OP_OR(result, A, x); return(result); }
    friend booleanMatrix operator||(CONST Element x, CONST Matrix& B)  	// x || B
      { booleanMatrix result(B.shape); OP_OR(result, x, B); return(result); }
    friend booleanMatrix operator!(CONST Matrix& A)                     // !A
      { booleanMatrix result(A.shape); OP_NOT(result, A); return(result); }
#endif // BOOLEAN_ELEMENTS

    friend ostream& operator<<(ostream&, CONST Matrix&);     // print A
    friend istream& operator>>(istream&,       Matrix&);     // read A
    
    // scans
#ifdef BOOLEAN_ELEMENTS			    // only booleans support OR and AND
    friend Matrix or_scan_rows(CONST Matrix&);   // or-scan each row
    friend Matrix or_scan_cols(CONST Matrix&);   // or-scan each col
    friend Matrix and_scan_rows(CONST Matrix&);  // and-scan each row
    friend Matrix and_scan_cols(CONST Matrix&);  // and-scan each col
#else
    friend Matrix plus_scan_rows(CONST Matrix&); // plus-scan each row
    friend Matrix plus_scan_cols(CONST Matrix&); // plus-scan each col
    friend Matrix max_scan_rows(CONST Matrix&);  // max-scan each row
    friend Matrix max_scan_cols(CONST Matrix&);  // max-scan each col
    friend Matrix min_scan_rows(CONST Matrix&);  // min-scan each row
    friend Matrix min_scan_cols(CONST Matrix&);  // min-scan each col
#endif

    // reductions
#ifndef BOOLEAN_ELEMENTS		   // booleans don't support max and min
//  friend int max_index(CONST Matrix&);   // index of max value in Matrix
//  friend int min_index(CONST Matrix&);   // index of min value in Matrix
#endif

    // shuffling around
//  friend Matrix permute(CONST Matrix& M, CONST Matrix& P); // permute M by P
//  friend Matrix permute(CONST Matrix& M, int (*f)(CONST int index)); //M by f
//  friend Matrix pack(CONST Matrix& M);		 // non-zeros go left

    friend Matrix shift(CONST Matrix& M, CONST int dr, CONST int dc);
    friend Matrix shift_rows(CONST Matrix& M, CONST intVector& cv); // shift each row differently
    friend Matrix shift_rows(CONST Matrix& M, CONST int& offset) { return(shift(M, 0, offset)); }
    friend Matrix shift_cols(CONST Matrix& M, CONST intVector& rv); // shift each col differently
    friend Matrix shift_cols(CONST Matrix& M, CONST int& offset) { return(shift(M, offset, 0)); }

    friend Matrix rotate(CONST Matrix& M, CONST int dr, CONST int dc);
    friend Matrix rotate_rows(CONST Matrix& M, CONST intVector& cv); // rotate each row differently
    friend Matrix rotate_rows(CONST Matrix& M, CONST int& offset) { return(rotate(M, 0, offset)); }
    friend Matrix rotate_cols(CONST Matrix& M, CONST intVector& rv); // rotate each col differently
    friend Matrix rotate_cols(CONST Matrix& M, CONST int& offset) { return(rotate(M, offset, 0)); }

    // miscellaneous public functions
    void load(CONST Element **CONST array);	 // load matrix from array
    void store(Element **CONST array) CONST; // store matrix into array
#ifndef BOOLEAN_ELEMENTS	// disallow arithmetic on booleans
    friend Matrix outer(CONST Vector& A, CONST Vector& B); // outer product A * B
#endif
    friend Matrix apply(Element (*f)(CONST Element), CONST Matrix& M) // f(M)
      {	Matrix result(M.shape); apply(result, f, M); return(result); }

    // row slices
    Slice operator[](CONST int index) {
	INVARIANT3(index >= 0 && index < shape.dim1, 
		   "Sorry, you can't get row [%d] from this matrix.", index);
	return (Slice(rowSlice, index, this));
    }

    // column slices; we create a half slice, which must be subscripted again
    HalfSlice operator[](CONST SliceDummy&) {
	return (HalfSlice(this));
    }

private: 
    // see other init functions used from Collection
    // initialize the matrix from a function of the index
    void init(Element (*f)(CONST int row, CONST int col));
    // initialize the matrix from an array
    void init(CONST Element **data); // Element data[rows][cols], we assume

    // these are used by Slice, a friend of ours
    Vector extract(CONST SliceType type, CONST int index) CONST;
    void update(CONST SliceType type, CONST int index, CONST Vector& V);

    // note we inherit operator()(CONST int r, CONST int c)  from Collection.
};

// to do:
// () function call could have some meaning?
// rank, sort, other NESL stuff?
