// @TITLE "Slice.cc: class Slice"
//
// Slice class; included by {int,char,float,double,boolean}.cc.  
//
// A slice is created as a row or column
// from a matrix.  A Slice is a subclass of Vector, so this behaves like
// a vector for most operations.  However, we overload the assignment 
// operators, so that we can provide the illusion that this is really 
// a reference to the original matrix.  Thus, you can write expressions
// like
//        A.row(i) -= pivotValue * A.row(pivotRow);
//
// David Kotz 1994

// $Id: Slice.cc,v 1.14 95/12/17 00:08:02 dfk release Locker: dfk $

//////////////////////////////////////////
#ifdef BOOLEAN_ELEMENTS		// only define this once
// used for column slices [_][i]
CONST SliceDummy _;		// yes, that variable name is underscore '_'
#endif

//////////////////////////////////////////////////////////////////////
// define a conversion operator to convert a Slice into a Vector
Slice::operator Vector() CONST {
    return M->extract(type, rowcol);
}

//////////////////////////////////////////////////////////////////////
// Subscripting operators are used for second-subscripting a matrix.

// M[i][j] is an element reference; here, we are specifying the column index
Element& Slice::operator[](CONST int index) CONST {
    INVARIANT2(type == rowSlice,  
	       "You can't subscript a column slice, like M[_][i][j]!");
    INVARIANT4 (index >= 0 && index < M->shape.dim2,
		"Column index %d out of range [0..%d)", 
		index, M->shape.dim2);
    return ((*M)(rowcol, index)); // a reference to row rowcol, column index
}

// M[i][_] is the same as M[i]
Slice Slice::operator[](CONST SliceDummy&) CONST {
    INVARIANT2(type == rowSlice,  
	       "You can't subscript a column slice, like M[_][i][_]!");
    return *this;
}

// M[_][j] is a column slice
Slice HalfSlice::operator[](CONST int index) CONST {
    INVARIANT4 (index >= 0 && index < M->shape.dim2,
		"Column index %d out of range [0..%d)", 
		index, M->shape.dim2);
    return (Slice(colSlice, index, M));
}

//////////////////////////////////////////////////////////////////////
// Overload the assignment operators:
//    Note that M->update updates the matrix from a vector. 
// Note we cast the Slice into a Vector (using the above operator), 
// then call the vector-vector operator, returning a vector.

Slice& Slice::operator=(CONST Vector& B) { // A=B
    M->update(type, rowcol, B);
    return *this;
}

Slice& Slice::operator=(CONST Element x)  {	// A=x
    // build a vector of the right size, with all elements initialized to x
    Vector X(type == rowSlice ? M->shape.dim1 : M->shape.dim2, x);
    M->update(type, rowcol, X);
    return *this;
}

#ifndef BOOLEAN_ELEMENTS // no arithmetic on booleans

Slice& Slice::operator+=(CONST Vector& B) { // A+=B
    M->update(type, rowcol, Vector(*this) + B);
    return *this;
}

Slice& Slice::operator+=(CONST Element x)  { // A+=x
    M->update(type, rowcol, Vector(*this) + x);
    return *this;
}

Slice& Slice::operator-=(CONST Vector& B) { // A-=B
    M->update(type, rowcol, Vector(*this) - B);
    return *this;
}

Slice& Slice::operator-=(CONST Element x)  { // A-=x
    M->update(type, rowcol, Vector(*this) - x);
    return *this;
}

Slice& Slice::operator*=(CONST Vector& B) { // A*=B
    M->update(type, rowcol, Vector(*this) * B);
    return *this;
}

Slice& Slice::operator*=(CONST Element x)  { // A*=x
    M->update(type, rowcol, Vector(*this) * x);
    return *this;
}

Slice& Slice::operator/=(CONST Vector& B) { // A/=B
    M->update(type, rowcol, Vector(*this) / B);
    return *this;
}

Slice& Slice::operator/=(CONST Element x)  { // A/=x
    M->update(type, rowcol, Vector(*this) / x);
    return *this;
}

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
Slice& Slice::operator%=(CONST Vector& B) { // A%=B
    M->update(type, rowcol, Vector(*this) % B);
    return *this;
}

Slice& Slice::operator%=(CONST Element x)  { // A%=x
    M->update(type, rowcol, Vector(*this) % x);
    return *this;
}
#endif // INT_ELEMENTS

Slice Slice::operator++(void) { // ++A
    M->update(type, rowcol, ++Vector(*this));
    return *this;
}

Slice Slice::operator++(int zero)  { // A++
    M->update(type, rowcol, Vector(*this)++);
    return *this;
}

Slice Slice::operator--(void) { // --A
    M->update(type, rowcol, --Vector(*this));
    return *this;
}

Slice Slice::operator--(int zero)  { // A--
    M->update(type, rowcol, Vector(*this)--);
    return *this;
}

#endif // BOOLEAN_ELEMENTS

