// @TITLE "Vector.cc: class Vector"
//
//  Vector class; included by {int,char,float,double,boolean}.cc
//
// David Kotz 1994

// $Id: Vector.cc,v 2.16 95/12/17 00:07:58 dfk release Locker: dfk $

/////////////////////////////////////////////////////////////////////
// Other constructors and some operators are declared inline in Vector.h

Vector::Vector(CONST intVector& V) : Collection(V) {}
Vector::Vector(CONST charVector& V) : Collection(V) {}
Vector::Vector(CONST floatVector& V) : Collection(V) {}
Vector::Vector(CONST doubleVector& V) : Collection(V) {}
Vector::Vector(CONST booleanVector& V) : Collection(V) {}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Vector::init: private function to initialize vector"
//  This variant initializes the array to a function of the indices.
// Note that all entries are initialized, regardless of active set.

void Vector::init( Element (*f)(CONST int index) )
{
    // now initialize v[] using the function
    for (int i = 0; i < shape.dim1; i++)
      (*this)(i) = f(i);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Vector::init: private function to initialize vector"
//  Given an array of integers, initialize this object.
// Note that all entries are initialized, regardless of active set.

void Vector::init(CONST Element *array)
{
    // now initialize v[] using the array
    for (int i = 0; i < shape.dim1; i++)
      (*this)(i) = array[i];
}


/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Vector::load: load vector from array"
//    copy the plain array into the vector
//  the array better have be large enough

void Vector::load(CONST Element *CONST array)
{
    INVARIANT2(array != NULL, "NULL array passed to Vector::load.");
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    for (int i = 0; i < shape.dim1; i++)
      if (Active.isActive(i))
	(*this)(i) = array[i];
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Vector::store: store vector into array"
//    copy the vector into a plain integer array.
//  the array better have enough space

void Vector::store(Element *CONST array) CONST
{
    INVARIANT2(array != NULL, "NULL array passed to Vector::store.");
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < shape.dim1; i++)
      if (Active.isActive(i))
	array[i] = (*this)(i);
}


//////////////// Friends of the Vector class ///////////////////

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "or_scan(Vector)"
//  apply an orring scan to this vector, returning a new vector
// note this is a boolean or, not a bitwise or

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Vector or_scan(CONST Vector& V)
{
    Vector result(V.shape);	// uninitialized
    Element m = false;
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
	  result(i) = m;
	  m =  (m || V(i));
      } else
	result(i) = false;
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "and_scan(Vector)"
//  apply an anding scan to this vector, returning a new vector
// note this is a boolean and, not a bitwise and

#ifdef BOOLEAN_ELEMENTS		// only for booleans
Vector and_scan(CONST Vector& V)
{
    Vector result(V.shape);	// uninitialized
    Element m = true;
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
	  result(i) = m;
	  m = (m && V(i));
      } else
	result(i) = false;
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "plus_scan(Vector)"
//  apply a summing scan to this vector, returning a new vector

#ifndef BOOLEAN_ELEMENTS	// booleans don't support plus
Vector plus_scan(CONST Vector& V)
{
    Vector result(V.shape);	// uninitialized
    Element sum = 0;
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++) 
      if (Active.isActive(i)) {
	  result(i) = sum;
	  sum += V(i);
      } else
	result(i) = 0;
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "max_scan(Vector)"
//  apply a maxing scan to this vector, returning a new vector

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max and min
Vector max_scan(CONST Vector& V)
{
    Vector result(V.shape);	// uninitialized
    Element m = Smallest();
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++) 
      if (Active.isActive(i)) {
	  result(i) = m;
	  m = max(m, V(i));
      } else
	result(i) = 0;
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "min_scan(Vector)"
//  apply a minning scan to this vector, returning a new vector

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max and min
Vector min_scan(CONST Vector& V)
{
    Vector result(V.shape);	// uninitialized
    Element m = Biggest();
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
	  result(i) = m;
	  m = min(m, V(i));
      } else
	result(i) = 0;
    
    return result;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "first(Vector)"
//  return the value of the first (lowest VP#) active element
// return 0 if none are active.

Element first(CONST Vector& V)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	return V(i);

    return false;			// arbitrary return value if none are active
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "first_index(Vector)"
//  return the index of the first (lowest VP#) active element,
// or -1 if none are active.

int first_index(CONST Vector& V)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	return i;

    return -1;			// ie, none were active
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "max_index(Vector)"
//  return the position of the maximum value in the vector.
// if there is a tie, it returns the leftmost index.
// if the active set is empty, it returns -1.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max or min
int max_index(CONST Vector& V)
{
    Element m = Smallest();
    int pos = -1;
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	if (V(i) > m) {
	    m = V(i);
	    pos = i;
	}
    
    return pos;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "min_index(Vector)"
//  return the position of the minimum value in the vector.
// if there is a tie, it returns the leftmost index.
// if the active set is empty, it returns -1.

#ifndef BOOLEAN_ELEMENTS	// booleans don't support max or min
int min_index(CONST Vector& V)
{
    Element m = Biggest();
    int pos = -1;
    
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	if (V(i) < m) {
	    m = V(i);
	    pos = i;
	}
    
    return pos;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "permute(Vector,Vector): permute vector given permutation vector"
//   We are given an auxiliary vector that should contain a 1:1 permutation.
// That is, each element of the permutation vector should contain an index
// in [0..V.shape.dim1), and there should be no duplicates.  (We do not 
// verify that the permutation is 1:1, though we could someday.)  Element i of 
// vector V is stored in element permutation[i] of the output vector.

Vector permute(CONST Vector& V, CONST intVector& permutation)
{
    INVARIANT4(V.shape == permutation.shape, // check for conformity
	       "Shape mismatch: %s and %s", 
	       V.shape.name(), permutation.shape.name());
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    Vector result(V);		// initialize to same, in case not all active
    booleanVector used(V.shape.dim1, false);
    
    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
	  int destination = permutation(i);
	  
	  // check the permutation vector for validity and completeness
	  INVARIANT4 (destination >= 0 && destination < V.shape.dim1,
		      "Permutation destination %d out of range [0..%d)", 
		      destination, V.shape.dim1);
	  INVARIANT2(!used(destination), "vector provided for permute() is mapping two things to the same destination");
	  used(destination) = true;
	  
	  result(permutation(i)) = V(i);
      }
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "permute(Vector,func): permute one vector given permutation function"
//   We are given a function f() that should compute a 1:1 permutation.
// That is, f() takes a parameter index in [0..V.shape.dim1), and returns 
// the destination index in [0..V.shape.dim1).   (We do not verify
// that the permutation is 1:1, though we could someday.)  Element i of 
// vector V is stored in element permutation[f(i)] of the output vector.

Vector permute(CONST Vector& V, int (*f)(CONST int index))
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    Vector result(V);		// initialize to same, in case not all active
    booleanVector used(V.shape.dim1, false);
    
    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
	  int destination = f(i);

	  // check the permutation vector for validity and completeness
	  INVARIANT4 (destination >= 0 && destination < V.shape.dim1,
		      "Permutation destination %d out of range [0..%d)", 
		      destination, V.shape.dim1);
	  INVARIANT2(!used(destination), "function provided for permute() is mapping two things to the same destination");
	  used(destination) = true;

	  result(destination) = V(i);
      }
    
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "pack(Vector): pack the vector"
//   Pack the vector by squishing all non-zeros to the left, and all zeros
// to the right.  The new vector is returned.

Vector pack(CONST Vector& V)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    int nz = 0;
    Vector result(V.shape);
    result = 0;			// initially all zeros
    
    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	if (V(i) != 0)
	  result(nz++) = V(i);
    
    return result;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "shift(Vector,distance): shift the vector"
//   Shift the given vector by the given offset.  Any positive or negative
// offset will do, but anything larger than +-V.shape.dim1 is silly.
// Positive offsets push things to the right (toward larger indices).
// The vacated spaces are filled with zeroes (may want more flexibility later).
// Inactive places in the vector are untouched. 
//   Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole vector is shifted delta places
// to the right.  [Alternatively, it could shift only the active elements,
// or shift them among the active positions.  So far, I've
// found the semantics here to be more useful.]

Vector shift(CONST Vector& V, CONST int delta)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    if (delta == 0)
      return V;		// zero shift is a no-op
    else  {
	Vector result(V.shape);
	result = 0; // initialize to 0
	
	int imin = max(0,          0          - delta);
	int imax = min(V.shape.dim1, V.shape.dim1 - delta);

	// now fill in those places that are not zero
	// move position i to position i+delta
	for (int i = imin; i < imax; i++)
	  result(i + delta) = V(i);

	return result;
    }
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "rotate(Vector,distance): rotate the vector"
//   Rotate the given vector by the given offset.  Any positive or negative
// offset will do, but anything larger than +-V.shape.dim1 is redundant.
// Positive offsets push things to the right (toward larger indices).
//   Note that (unlike most functions) this function ignores the active set,
// except to check conformity.  Thus, the whole vector is rotated deltas 
// places to the right.  [Alternatively, it could rotate only the active 
// elements, among the active positions.  So far, I've
// found the semantics here to be more useful.]

Vector rotate(CONST Vector& V, CONST int delta)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    int offset = delta;		// local, changable copy

    offset %= V.shape.dim1;	// eliminate extra full rotations
    offset += V.shape.dim1;	// convert negative to positive
    offset %= V.shape.dim1;	// ie, now it is always a right rotation
    
    if (offset == 0)
      return V;			// zero rotation is a no-op
    else {
	Vector result(V.shape);	// uninitialized

	// move position i to position i+1, mod the vector size
	for (int i = 0; i < V.shape.dim1; i++)
	  result((i + offset) % V.shape.dim1) = V(i);
	
	return result;
    }
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "ostream operator <<: print operator"
//   Print vector A to the given output stream.
// No newline is appended;  elements are tab-separated.

ostream& operator<< (ostream& out, CONST Vector& V)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

#ifndef CHAR_ELEMENTS
    boolean first = true;
#endif // CHAR_ELEMENTS
    
    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i)) {
#ifndef CHAR_ELEMENTS
	  if (first)
	    first = false;
	  else
	    out << "\t";
#endif // CHAR_ELEMENTS

	  out << V(i);
      }
    return(out);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "istream operator >>: read operator"
//   Read vector V from the given input stream.
// No newline is expected;  elements should be whitespace-separated.

istream& operator>> (istream& in, Vector& V)
{
    INVARIANT4(Active.shapeOK(V.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       V.shape.name(), Active.shapeName());

    for (int i = 0; i < V.shape.dim1; i++)
      if (Active.isActive(i))
	in >> V(i);
    return(in);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "inner(): inner product of two vectors"
// compute inner product A * B
// A and B should have the same number of elements; the result is an element.
// Only those locations that correspond to active VPs will be included in the product.
// If none are active then 0 is returned.

#ifndef BOOLEAN_ELEMENTS	// disallow arithmetic on booleans
Element inner(CONST Vector& A, CONST Vector& B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());
    
    Element sum = 0;

    for (int i = 0; i < A.shape.dim1; i++)
      if (Active.isActive(i))
	sum += A(i) * B(i);

    return(sum);
}
#endif

/////////////////////////////////////////////////////////////////////
