// @TITLE "Vector.h: class Vector"
//
// included by dapple.h -- do not include this directly!
//
// Vector class.  A Vector is a subclass of Collection, so we can inherit
// many of the elementwise operators.  Thus, we need only add functions that
// make sense on Vectors; in particular, functions that involve the
// understanding that the elements of the collection have some order, 
// and each has some index.  This includes some scan operators, some 
// order-dependent reductions, permutations,
// input and output, and some new constructors to initialize from an array or
// a function of the index.
//
// David Kotz 1994

// $Id: Vector.h,v 2.16 95/12/17 00:08:00 dfk release Locker: dfk $

class Vector : public Collection {

    friend class Matrix; // so it can work with slices, which are vectors

    // they're all friends
#if !defined(INT_ELEMENTS)
    friend class intVector;
#endif
#if !defined(CHAR_ELEMENTS)
    friend class charVector;
#endif
#if !defined(FLOAT_ELEMENTS)
    friend class floatVector;
#endif
#if !defined(DOUBLE_ELEMENTS)
    friend class doubleVector;
#endif
#if !defined(BOOLEAN_ELEMENTS)
    friend class booleanVector;
#endif

public: 
    // constructors and destructors
    Vector(CONST int size)
      : Collection(Shape('V', size)) {}
    Vector(CONST int size, CONST Element x) 
      : Collection(Shape('V', size), x) {}
    Vector(CONST int size, CONST Element *d)
      : Collection(Shape('V', size)) { init(d); }
    Vector(CONST int size, Element (*f)(CONST int index))
      : Collection(Shape('V', size)) { init(f); }
    Vector(CONST intVector& V);	// similar to the above
    Vector(CONST charVector& V); // but defined in Vector.cc
    Vector(CONST floatVector& V);
    Vector(CONST doubleVector& V);
    Vector(CONST booleanVector& V);
    
public:		// internal; should be private! but strictly speaking won't work that way.
    Vector(CONST Shape& s) : Collection(s) {}
    
public:
    // overloaded operators
    Vector& operator=(CONST Vector& B)   // A = B
      { this->OP_Equals(B); return *this; }
    Vector& operator=(CONST Element x)   // A = x
      { this->OP_Equals(x); return *this; }
#ifndef BOOLEAN_ELEMENTS   // don't allow arithmetic on booleans
    Vector& operator+=(CONST Vector& B)   // A += B
      { this->OP_PlusEquals(B); return(*this); }
    Vector& operator+=(CONST Element x)   // A += x
      { this->OP_PlusEquals(x); return(*this); }
    Vector& operator-=(CONST Vector& B)   // A -= B
      { this->OP_MinusEquals(B); return(*this); }
    Vector& operator-=(CONST Element x)   // A -= x
      { this->OP_MinusEquals(x); return(*this); }
    Vector& operator*=(CONST Vector& B)   // A *= B
      { this->OP_TimesEquals(B); return(*this); }
    Vector& operator*=(CONST Element x)   // A *= x
      { this->OP_TimesEquals(x); return(*this); }
    Vector& operator/=(CONST Vector& B)   // A /= B
      { this->OP_DivideEquals(B); return(*this); }
    Vector& operator/=(CONST Element x)   // A /= x
      { this->OP_DivideEquals(x); return(*this); }
# ifdef INT_ELEMENTS // not for char, float, double, or boolean
    Vector& operator%=(CONST Vector& B)   // A %= B
      { this->OP_ModEquals(B); return(*this); }
    Vector& operator%=(CONST Element x)   // A %= x
      { this->OP_ModEquals(x); return(*this); }
# endif
    Vector  operator++(void)  	         // ++A
      { this->OP_PlusPlus(); return(*this); }
    Vector  operator++(int zero)   // A++
      { Vector result(*this); this->OP_PlusPlus(zero); return(result); }
    Vector  operator--(void)  	         // --A
      { this->OP_MinusMinus(); return(*this); }
    Vector  operator--(int zero)   // A--
      { Vector result(*this); this->OP_MinusMinus(zero); return(result); }

    friend Vector       operator+(CONST Vector& A)  	  	      // +A 
      {	return (A); }
    friend Vector       operator-(CONST Vector& A) 		       // -A 
      { Vector result(A.shape);	OP_Minus(result, A); return(result); }
    friend Vector       operator+(CONST Vector& A, CONST Vector& B)   // A + B
      {	Vector result(A.shape); OP_Plus(result, A, B); return(result); }
    friend Vector       operator+(CONST Vector& A, CONST Element x)   // A + x
      {	Vector result(A.shape); OP_Plus(result, A, x); return(result); }
    friend Vector       operator+(CONST Element x, CONST Vector& B)   // x + B
      {	Vector result(B.shape); OP_Plus(result, x, B); return(result); }
    friend Vector       operator-(CONST Vector& A, CONST Vector& B)   // A - B
      { Vector result(A.shape); OP_Minus(result, A, B); return(result); }
    friend Vector       operator-(CONST Vector& A, CONST Element x)   // A - x
      { Vector result(A.shape); OP_Minus(result, A, x); return(result); }
    friend Vector       operator-(CONST Element x, CONST Vector& B)   // x - B
      { Vector result(B.shape); OP_Minus(result, x, B); return(result); }
    friend Vector       operator*(CONST Vector& A, CONST Vector& B)   // A * B
      { Vector result(A.shape); OP_Times(result, A, B); return(result); }
    friend Vector       operator*(CONST Vector& A, CONST Element x)   // A * x
      { Vector result(A.shape); OP_Times(result, A, x); return(result); }
    friend Vector       operator*(CONST Element x, CONST Vector& B)   // x * B
      { Vector result(B.shape); OP_Times(result, x, B); return(result); }
    friend Vector       operator/(CONST Vector& A, CONST Vector& B)   // A / B
      { Vector result(A.shape); OP_Divide(result, A, B); return(result); }
    friend Vector       operator/(CONST Vector& A, CONST Element x)   // A / x
      { Vector result(A.shape); OP_Divide(result, A, x); return(result); }
    friend Vector       operator/(CONST Element x, CONST Vector& B)   // x / B
      { Vector result(B.shape); OP_Divide(result, x, B); return(result); }
# ifdef INT_ELEMENTS			       // not for char, float, double, or boolean
    friend Vector       operator%(CONST Vector& A, CONST Vector& B)   // A % B
      { Vector result(A.shape); OP_Mod(result, A, B); return(result); }
    friend Vector       operator%(CONST Vector& A, CONST Element x)   // A % x
      { Vector result(A.shape); OP_Mod(result, A, x); return(result); }
    friend Vector       operator%(CONST Element x, CONST Vector& B)   // x % B
      { Vector result(B.shape); OP_Mod(result, x, B); return(result); }
# endif // INT_ELEMENTS

    friend booleanVector operator<(CONST Vector& A, CONST Vector& B)   // A < B
      { booleanVector result(A.shape); OP_LT(result, A, B); return(result); }
    friend booleanVector operator<(CONST Vector& A, CONST Element x)   // A < x
      { booleanVector result(A.shape); OP_LT(result, A, x); return(result); }
    friend booleanVector operator<(CONST Element x, CONST Vector& B)   // x < B
      { booleanVector result(B.shape); OP_LT(result, x, B); return(result); }
    friend booleanVector operator>(CONST Vector& A, CONST Vector& B)   // A > B
      { booleanVector result(A.shape); OP_GT(result, A, B); return(result); }
    friend booleanVector operator>(CONST Vector& A, CONST Element x)   // A > x
      { booleanVector result(A.shape); OP_GT(result, A, x); return(result); }
    friend booleanVector operator>(CONST Element x, CONST Vector& B)   // x > B
      { booleanVector result(B.shape); OP_GT(result, x, B); return(result); }
    friend booleanVector operator<=(CONST Vector& A, CONST Vector& B)   // A <= B
      { booleanVector result(A.shape); OP_LE(result, A, B); return(result); }
    friend booleanVector operator<=(CONST Vector& A, CONST Element x)  	// a <= x
      { booleanVector result(A.shape); OP_LE(result, A, x); return(result); }
    friend booleanVector operator<=(CONST Element x, CONST Vector& B)  	// x <= B
      { booleanVector result(B.shape); OP_LE(result, x, B); return(result); }
    friend booleanVector operator>=(CONST Vector& A, CONST Vector& B)   // A >= B
      { booleanVector result(A.shape); OP_GE(result, A, B); return(result); }
    friend booleanVector operator>=(CONST Vector& A, CONST Element x)  	// A >= x
      { booleanVector result(A.shape); OP_GE(result, A, x); return(result); }
    friend booleanVector operator>=(CONST Element x, CONST Vector& B)  	// x >= B
      { booleanVector result(B.shape); OP_GE(result, x, B); return(result); }
#endif // BOOLEAN_ELEMENTS
    friend booleanVector operator==(CONST Vector& A, CONST Vector& B)   // A == B
      { booleanVector result(A.shape); OP_EQ(result, A, B); return(result); }
    friend booleanVector operator==(CONST Vector& A, CONST Element x)  	// A == x
      { booleanVector result(A.shape); OP_EQ(result, A, x); return(result); }
    friend booleanVector operator==(CONST Element x, CONST Vector& B)  	// x == B
      { booleanVector result(B.shape); OP_EQ(result, x, B); return(result); }
    friend booleanVector operator!=(CONST Vector& A, CONST Vector& B)   // A != B
      { booleanVector result(A.shape); OP_NE(result, A, B); return(result); }
    friend booleanVector operator!=(CONST Vector& A, CONST Element x)  	// A != x
      { booleanVector result(A.shape); OP_NE(result, A, x); return(result); }
    friend booleanVector operator!=(CONST Element x, CONST Vector& B)  	// x != B
      { booleanVector result(B.shape); OP_NE(result, x, B); return(result); }

#ifdef BOOLEAN_ELEMENTS   // only for booleans
    friend booleanVector operator&&(CONST Vector& A, CONST Vector& B)   // A && B
      { booleanVector result(A.shape); OP_AND(result, A, B); return(result); }
    friend booleanVector operator&&(CONST Vector& A, CONST Element x)  	// A && x
      { booleanVector result(A.shape); OP_AND(result, A, x); return(result); }
    friend booleanVector operator&&(CONST Element x, CONST Vector& B)  	// x && B
      { booleanVector result(B.shape); OP_AND(result, x, B); return(result); }
    friend booleanVector operator||(CONST Vector& A, CONST Vector& B)   // A || B
      { booleanVector result(A.shape); OP_OR(result, A, B); return(result); }
    friend booleanVector operator||(CONST Vector& A, CONST Element x)  	// A || x
      { booleanVector result(A.shape); OP_OR(result, A, x); return(result); }
    friend booleanVector operator||(CONST Element x, CONST Vector& B)  	// x || B
      { booleanVector result(B.shape); OP_OR(result, x, B); return(result); }
    friend booleanVector operator!(CONST Vector& A)                     // !A
      { booleanVector result(A.shape); OP_NOT(result, A); return(result); }
#endif // BOOLEAN_ELEMENTS

    friend ostream& operator<<(ostream&, CONST Vector&);     // print A
    friend istream& operator>>(istream&,       Vector&);     // read A
    Element& operator[](CONST int index) CONST { return v[index]; } // A[index]
    
    // scans
#ifdef BOOLEAN_ELEMENTS			    // only booleans support OR and AND
    friend Vector or_scan(CONST Vector&);   // do an or-scan and return new vector
    friend Vector and_scan(CONST Vector&);  // do an and-scan and return new vector
#else					    // booleans don't support plus, max, or min
    friend Vector plus_scan(CONST Vector&); // do a plus-scan and return new Vector
    friend Vector max_scan(CONST Vector&);  // do a max-scan and return new vector
    friend Vector min_scan(CONST Vector&);  // do a min-scan and return new vector
#endif

    // reductions
    friend Element first(CONST Vector&);   // return value of first active element
    friend int first_index(CONST Vector&); // return index of first active element
#ifndef BOOLEAN_ELEMENTS		   // booleans don't support max and min
    friend int max_index(CONST Vector&);   // return index of max value in vector
    friend int min_index(CONST Vector&);   // return index of min value in vector
#endif

    // shuffling around
    friend Vector permute(CONST Vector& V, CONST intVector& P); // permute V by P
    friend Vector permute(CONST Vector& V, int (*f)(CONST int index));// V by f
    friend Vector pack(CONST Vector& V);		 // non-zeros go left
    friend Vector shift(CONST Vector& V, CONST int distance); // shift V
    friend Vector rotate(CONST Vector& V, CONST int distance); // rotate V

    // miscellaneous public functions
    void load(CONST Element *CONST array);	 // load vector from array
    void store(Element *CONST array) CONST;	 // store vector into array
#ifndef BOOLEAN_ELEMENTS	// disallow arithmetic on booleans
    friend Element inner(CONST Vector& A, CONST Vector& B); // inner product, ie, dot product
#endif
    friend Vector apply(Element (*f)(CONST Element), CONST Vector& V)   // f(V)
      { Vector result(V.shape); apply(result, f, V); return(result); }

private: 
    // see other init functions used from Collection
    // initialize the vector from a function of the index
    void init(Element (*f)(CONST int index));
    // initialize the vector from an array
    void init(CONST Element *data);	// Element data[size], we assume
      
    // note we inherit operator()(CONST int i)  from Collection.
};
